"Python3 program to split time-series into sections."

##  EASTIRS :
##
##  Evolutionary Algorithm Slicing Timeseries Into Related Segments :
##  Copyright 2014, R.S. Forsyth.
##  first version : 19/08/2014 (26/03/1994 in C)
##  last revision : 23/10/2014


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


import  math, os, string, sys
import  random, time


##  local libraries :
from globdefs import *
from glib import  numeric, numerize
from glib import  parafind, paraget
from glib import  parashow
from glib import  showobj
from glib import  openfile
from glib import  request
from gridlib import  cols2use, matread
from gridlib import  matdump, vecline
from gridlib import  var2col
from gridlib import  bastatz, midvalue
from py3seal import Genepool


##  parameter info :
##  List of parameters that need 2b set, with defaults & lo,hi limits if appropriate :
Paralist = [
    ("atomize",1,0,1),
    ("filextn",''),
    ("targcol",0,0),  ##  column 2use as outcome
    ("targvar",''),  ##  name of outcome/target variable
    ("timevar",''),  ##  name of time-stamp variable
    ("ntrials",49152,256,262144),  ##  number of evolutionary trials
    ("varmode","vc"),  ##  variance mode, from : v1,v2,vc,vx (vc works best)
    ("jobname","eastirs"),  ##  name to link related files
    ("casefold",1,0,1),  ##  whether to fold to lower case
    ("datfile",''),  ##  file for input of data
    ("skipvars",''),  ##  variables to ignore
    ("showgap",4096,256,262144),  ##  mingap between flea() displays
    ("listfile",''),  ##  main output file
    ("dumpfile",''),  ##  output file for derived data
    ("outpath",os.pardir+os.sep+"op"),  ##  default o/p drive/path
    ("outfile",os.pardir+os.sep+"op"+os.sep+"eastirs.txt"),  ##  logfile 4checking
    ("zonk",0,0,1)  ##  just a sentinel
    ]
##  Each item above is a tuple: name,default,minvalue,maxvalue.
##  Contents of parafile will take precedence over these defaults,
##  but parameters not in Paralist won't be used even if in file.
##  0/1 used as True/False (lazy but easy!).
##  parafile syntax, line by line:
##  paraname <blank(s)> paraval


##  global vars :
Version = "4.0"
##  Gold = 5.0**0.5*0.5+0.5
Mincases = 4
Dumptail = "_dump"
Listtail = "_list"
Paratail = "_pars"
Intname = "integers"
Popsize = 233
Gmax  = 1  ##  maximum gene integer (0..Gmax)


##  Main program :
##
def main():
    "Main line."

    ##  First some odds & ends concerned with orientation :
    args = sys.argv[:]          ##  gets command-line arguments
    progname = args[0]
    dateline = time.asctime(time.localtime(time.time()))
    paraname, prepaths = parafind(args,dateline,Version)
    startime = time.time()

    if paraname:
        parafile = openfile(paraname,prepaths,[".pf",".par",".txt",'.'],'r')
    else:
        parafile = None
    ##  print(parafile)
    if parafile == None:
        print("Cannot proceed without a parameter file!",NL,"Program halts.")
        return  -1

    ##  Read parameter settings :
    paradat = paraget(Paralist,parafile)
    ##  print(list(paradat.__dict__.items()),NL)
    parafile.close()
    
    ##  do some tidying of paradat :
    paradat.dateline = dateline
    paradat.progname = progname
    if hasattr(paradat,"casefold"):
        paradat.foldcase = paradat.casefold
    elif hasattr(paradat,"foldcase"):
        paradat.casefold = paradat.foldcase
    if hasattr(paradat,"filextn"):
        if len(paradat.filextn) > 0 and not paradat.filextn.startswith('.'):
            paradat.filextn = '.' + paradat.filextn
            ##  ensures initial fullstop
    else:
        paradat.filextn = ''  ##  must have filextn, even if empty
        ##  empty string means read any file type
    if not paradat.outpath.endswith(os.sep):
        paradat.outpath = paradat.outpath + os.sep
        ##  must have final separator character.
        if not os.path.isdir(paradat.outpath):  ##  try another outpath
            paradat.outpath = os.getcwd() + os.sep + "op" + os.sep
            if not os.path.isdir(paradat.outpath):
                os.makedirs(paradat.outpath)  ##  create o/p folder
            paradat.outfile = paradat.outpath + os.path.basename(paradat.outfile)
            
    paradat.datamat,paradat.taglist,paradat.tagdict = matread(paradat.datfile,
                                                              paradat)
    paradat.rows = len(paradat.datamat)
    print("data rows =",paradat.rows)
    paradat.cols = len(paradat.taglist)
    print("data cols =",paradat.cols)
    print(paradat.taglist)
    if paradat.rows < Mincases:  ##  not enough data
        print("Not enough data cases !")
        print("Program halts.")
        return  -1
    goodcols = cols2use(paradat)
    paradat.goodcols = goodcols
    if len(goodcols) < 1:
        print("No numeric variables found in ",paradat.datfile,'!')
        print("Program halts.")
        return  -2
    ##  obtain time-stamp variable :
    print("Usable variables : ")
    for v in sorted(paradat.goodcols):
        print(v,paradat.taglist[v])
    gott = False
    if paradat.timevar and paradat.timevar in paradat.taglist:
        tpos = var2col(paradat.timevar,paradat.taglist,goodcols)
        if tpos >= 0:
            gott = True  ##  given in parameter file
            varname = paradat.timevar
    while not gott:
        varname = request("Please give name of time-stamp variable",'')
        tpos = var2col(varname,paradat.taglist,goodcols)
        if tpos < 0:  ##  didn't match
            print(varname,"not found in usable variable names!")
            mess = "Do you want to use 1.."+str(paradat.rows)+" as time-stamp"
            tz = request(mess,'y')
            if tz == 'y':
                gott = True
                tpos = paradat.cols ; varname = Intname
            else:  ##  just a mistyping, retry
                print("Available column names :")
                for v in range(paradat.cols):
                    if v in goodcols:
                        print(paradat.taglist[v])
        else:  ##  okay
            tcol = tpos ; gott = True
    if tpos == paradat.cols:  ##  have 2use positive integers as time-stamp
        tvec = list(range(1,paradat.rows+1))
    else:
        print("Time-stamp chosen is column",tpos,paradat.taglist[tpos])
        tvec = [paradat.datamat[j][tpos] for j in range(paradat.rows)]
    if tpos < paradat.cols:
        paradat.timevar = paradat.taglist[tpos]  ##  save 4later
    else:
        paradat.timevar = Intname

    ##  print(goodcols)
    print(paradat.timevar)
    print(tvec[:10]) ; print("....") ; print(tvec[-10:])

    ##  next obtain y-variable name :
    goodcols.discard(tpos)
    goty = False
    while not goty:
        paradat.targcol = var2col(paradat.targvar,paradat.taglist,goodcols)
        if paradat.targcol in goodcols:
            print("time-series y-variable :")
            print(paradat.targcol,paradat.targvar)
            yvec = [paradat.datamat[j][paradat.targcol] for j in range(paradat.rows)]
            goty = True
        else:
            if paradat.targvar:
                print("y-variable",paradat.targvar,"not suitable.")
            else:
                print("y-variable needed.")
            print("usable variables :")
            for v in range(paradat.cols):
                if v in paradat.goodcols:
                    print(v,paradat.taglist[v])
            paradat.targvar = request("Please give name of y-variable",'')            

    ##  compute basic stats of target variable :
    paradat.bs = bastatz(yvec)
    print("basic stats of",paradat.targvar,":")
    showobj(paradat.bs)
    ##  check for degenerate input :
    if paradat.bs.var1 <= Tiny:  ##  no variance
        print("Input y-variable has zero variance!")
        print("Program cannot continue.")
        return  -3
    
    ##  extract & link up t & y vars :
    triplist = []  ##  will hold triples of : tval, rowindex, yval
    for r in range(paradat.rows):
        triplist.append((tvec[r],r,yvec[r]))
    triplist.sort()  ##  ensure ordering (stable because of r)
    ##  for r in range(paradat.rows):
        ##  triplist[r] = (triplist[r][0],triplist[r][-1])  ##  remove row number
    ##  print(triplist)
    paradat.triplist = triplist

    ##  fill step-up vector ??

    ##  apply evolutionary optimzer :
    paradat.gmax = Gmax
    paradat.glen = paradat.rows
    paradat.lastshow = 0
    ##  mu = Gold*42/paradat.glen
    mu = Gold
    pool = Genepool("eastirs1",fitness,paradat,paradat.glen,paradat.gmax,mu)
    bias = []
    ##  execute evolutionary optimization :
    pool.initpop(Popsize,bias)
    pool.flea(paradat.ntrials)
    print(pool.best)
    bi = pool.bestitem
    print(bi)

    ##  prepare o/p files :
    if paradat.outfile:
        fo = open(paradat.outfile,'w',encoding="utf-8")
    else:
        outname = paradat.outpath + paradat.jobname + Paratail + ".txt"
        fo = open(outname,'w',encoding="utf-8")
    if paradat.dumpfile:
        dumpname = paradat.dumpfile
    else:
        dumpname = paradat.outpath + paradat.jobname + Dumptail + ".dat"
    dumpfile = open(dumpname,'w',encoding="utf8")
    if paradat.listfile:
        listname = paradat.listfile
    else:
        listname = paradat.outpath + paradat.jobname + Listtail + ".txt"
    listfile = open(listname,'w',encoding="utf8")    

    ##  o/p of results :
    parashow(paradat,fo) ; fo.close()
    paradat.seglist = pout(listfile,paradat,pool.best,pool.bestitem)
    ##  seglist is an extra column of group numbers.
    fullname = os.path.abspath(listfile.name)
    print("output listing on file ",fullname)
    listfile.close()
    fullname = os.path.abspath(dumpfile.name)
    print("derived data dumped on ",fullname)
    dumpout(dumpfile,paradat,pool.best,pool.bestitem)
    dumpfile.close()
    
    print(progname,"done on",time.asctime(time.localtime(time.time())))
    print("after",round(time.time()-startime,5),"seconds.")  ##  secs since kickoff
    return  0




##  DUMPOUT :
def dumpout (dumpfile,paradat,bestval,bi,dlim=Htab):
    "writes extended data file to dumpfile, delimited with headline."

    headline = dlim.join(paradat.taglist)
    headline += dlim + "seg_ment"  ##  last column is segment code
    print(headline,file=dumpfile)

    ptab = {}
    ##  restore to original ordering :
    for j,trip in enumerate(paradat.triplist):
        tval,r,yval = trip
        ptab[r] = j
    ##  dump with extra group-code field :
    for j in range(len(ptab)):
        rr = ptab[j]
        trip = paradat.triplist[rr]
        ##  print(j,rr,trip)
        tval,r,yval = trip
        datline = paradat.datamat[r]  ##  input data case
        datline.append(paradat.seglist[rr])  ##  extra field
        datline = [str(d) for d in datline]
        datline = dlim.join(datline)
        print(datline,file=dumpfile)       

    return


##  POUT :
def pout (listfile,paradat,bestval,bi):
    "writes evolutionary optimizer's results to already-opened listfile."

    ##  header info :
    print(paradat.progname,file=listfile)
    print(paradat.dateline,file=listfile)
    print("\nSource file name :",paradat.datfile,file=listfile)
    print("Sequence variable :",paradat.timevar,file=listfile)
    print("Response variable :",paradat.targvar,file=listfile)
    ybar = paradat.bs.mean ; var2 = paradat.bs.var2
    print(paradat.targvar,"mean =",round(ybar,7),file=listfile)
    print(paradat.targvar,"variance =",round(var2,7),file=listfile)

    print("\nFitness mode :",paradat.varmode,file=listfile)
    print("Best score = ",round(bestval,7),file=listfile)
    print("\nBest partition :",file=listfile)
    print("Rank  Case  Gene  Group ",paradat.timevar[:12].rjust(12),' ',
          paradat.targvar[:12].rjust(12),file=listfile)
    g = 0
    segment = [] ; segs = []
    seglist = []
    locs = []
    for i,b in enumerate(bi):
        if b > 0 and segment:
            ##  print(segment,file=listfile)
            segs.append(segment)
        tval,r,yval = paradat.triplist[i]
        if b > 0:
            g += 1
            segment = [yval]
            locs.append(tval)
        else:
            segment.append(yval)
        seglist.append(g)  ##  group code of every item in triplist
        lout = "{0:4d}  {1:4d}  {2:4d}  {3:5d}".format(i+1,r+1,b,g)
        lout += "  {0:12.4f}   {1:12.4f}".format(tval,yval)
        print(lout,file=listfile)
        
    if segment:  ##  deal with last slice
        ##  print(segment,file=listfile)
        segs.append(segment)

    ##  also show segment stats :
    print("\nSegment statistics :",file=listfile)
    vtot = 0.0
    for g,segment in enumerate(segs):
        bx = bastatz(segment)
        ##  bx.id = g+1
        ##  showobj(bx,listfile)
        m = midvalue(segment)
        gbar = bx.mean ; gvar = bx.var2
        n = len(segment)
        if n < 2:
            gvar = 0.0
        lout = "\nSegment{0:4d} at {1:.4f}  n ={2:5d} :".format(g+1,locs[g],n)
        print(lout,file=listfile)
        lout = "median = {0:9.4f}  mean = {1:9.4f}  variance = {2:10.4f}".format(m,gbar,gvar)
        print(lout,file=listfile)
        vtot += gvar

    print("\nSummed subgroup variance =",round(vtot,7),file=listfile)

    return  seglist


##  VARIETY :
def variety (xvec,var1,var2,varmode="v2"):
    "computes variations on theme of variance."

    ##  var1 & var2 are global variances [/n & /(n-1)].
    n = len(xvec) ; frac = 1.0
    bx = bastatz(xvec)
    v1 = bx.var1 ; v2 = bx.var2  ##  local variances
    if n < 2:  ##  how best to deal with singletons ?
        v1 = var1*frac ; v2 = var2 * frac
    if varmode == "v1":
        vout = v1
    elif varmode == "v2":
        vout = v2
    elif varmode == "vx":
        w = 1.0 / n**2
        vout = (bx.var1 + var1*w) / (1.0+w)
    else:
        vout = v2

    return  vout


##  FITNESS :
def fitness (gvec,paradat):
    "fitness function for time-series slicer."

    if paradat.varmode == "vc":
        return  varcost(gvec,paradat)
    vtot = 0.0
    gvec[0] = 1  ##  redundant ??
    segment = [paradat.triplist[0][-1]]  ##  always start from front
    for g in range(1,paradat.glen):
        if gvec[g] > 0:  ##  new segment starts
            ##  deal with previous segment :
            v = variety(segment,paradat.bs.var1,paradat.bs.var2,paradat.varmode)
            vtot += v
            segment = [paradat.triplist[g][-1]]  ##  begin new segment
        else:  ##  continue with current segment
            segment.append(paradat.triplist[g][-1])
    ##  deal with last segment :
    if len(segment) > 0:
        vtot += variety(segment,paradat.bs.var1,paradat.bs.var2,paradat.varmode)
        
    return  1.0 - vtot / paradat.bs.var2
    ##  inverted for maximizer


##  VARCOST0 :
def varcost0 (gvec,paradat):
    "fitness function for time-series slicer."

    vtot = 0.0 ; var1 = paradat.bs.var1
    frac = 1.0  ##  1/Gold ??
    gvec[0] = 1  ##  redundant ??
    segment = [paradat.triplist[0][-1]]  ##  always start from front
    meanlist = []
    for g in range(1,paradat.glen):
        if gvec[g] > 0:  ##  new segment starts
            ##  deal with previous segment :
            bx = bastatz(segment)
            if len(segment) > 1:  ##  otherwise no variance
                vtot += bx.var1
            else:
                vtot += var1*frac
            meanlist.append(bx.mean)
            segment = [paradat.triplist[g][-1]]  ##  begin new segment
        else:  ##  continue with current segment
            segment.append(paradat.triplist[g][-1])
    ##  deal with last segment :
    if len(segment) == 1:  ##  singleton will have zero var1
        vtot += var1*frac
    elif len(segment) > 1:
        bx = bastatz(segment)
        vtot += bx.var1  ##  cost of each group
        meanlist.append(bx.mean)
    bx = bastatz(meanlist)  ##  cost having of groups
    if len(meanlist) > 1:
        gc = bx.var1
    elif len(meanlist) == 1:
        gc = var1 * frac
    
    return  -(gc + vtot)
    ##  negated for maximizer


##  VARCOST :
def varcost (gvec,paradat):
    "fitness function for time-series slicer."

    gvec[0] = 1  ##  redundant ??
    g = 0
    segment = [] ; segs = []
    for i,b in enumerate(gvec):
        if b > 0 and segment:
            segs.append(segment)
        tval,r,yval = paradat.triplist[i]
        if b > 0:
            g += 1
            segment = [yval]
        else:
            segment.append(yval)        
    if segment:  ##  deal with last slice
        segs.append(segment)
    ##  if g < 1 then error !!
    meanlist = [] ; etot = 0.0
    n = len(gvec)
    ##  compute within-group errors :
    for i,segment in enumerate(segs):
        ybar = sum(segment) / len(segment)
        d2 = [(ybar-y)**2 for y in segment]
        etot += sum(d2)  ##  cost of data
        meanlist.append(ybar)
    ##  between-group diffs :
    nm = len(meanlist)
    assert g == nm ; assert g == len(segs)
    gm = paradat.bs.mean  ##  grand mean
    mbar = sum(meanlist) / g
    d2 = sum([(gm-y)**2 for y in meanlist])  ##  cost of groups
    
    ##  return  paradat.bs.var2 -(d2 + etot) / max(1,n-g)
    return  1.0 - ((d2 + etot) / max(1,n-g)) / paradat.bs.var2
    ##  inverted for maximizer




##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
