"SEAL: Simple Evolutionary Algorithm Library, in Python3."

##  SEAL :
#
##  Richard's Simple Evolutionary Algorithm Library, module for importing.
##  Copyright 2000, R.S. Forsyth
##  first version : 22/11/2000
##  last revision : 10/01/2015

##  Currently assumes each genome is a list of integers.
##  N.B. fitness must be defined outside.

ShowGNU = 0  ##  make ShowGNU=0 when tired of seeing licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  Please
  see the GNU General Public License for more details. Thank you.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)


from copy import *
import math, os, string, sys, random


##  global vars :
NL = '\n'
alphanum = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


##  dummy fitness function :
def fitness (g, e):
    "dummy fitness function."
    ##  assumes g is a list of integers; e is environment.
    n = len(g)
    c = n // 2
    return g[0] * g[n-1] - g[c]


class Genepool:
    huge = 303e303
    tiny = 1e-36
    zonksize = 125
    showoff = 1  ##  exhibitionism parameter
    
    def __init__(self,name,fitfunc,context,glen=16,gmax=9,mupc=8.0,swapsize=0):
        self.name = name
        self.mu = mupc  ##  mutation rate as percentage
        self.glen = glen  ##  genome length
        self.genotype = []  ##  the population itself, a list
        self.quality = []  ##  fitness score for each item
        self.fitness = fitfunc  ##  fitness function
        self.context = context  ##  external environment object
        self.swapsize = swapsize  ##  gives info on blocking
        ##  fitness function must be defined outside.
        ##  context object will be problem-dependent.
        self.showoff = Genepool.showoff
        self.best = -Genepool.huge
        self.bestitem = None
        self.gmax = gmax  ##  maximum value at each gene position
        self.incr = int(round(math.log(gmax+1)))
        self.zonklist = [0,1]  ##  ensure at least a zero & 1
        self.lastshow = 0  ##  trial of last display
        self.ownshow = 1  ##  internal showbest() 2b used
        if hasattr(context,"topshow"):  ##  function to showoff best supplied
            self.topshow = context.topshow
            self.ownshow = 0  ##  external showbest supplied
        if hasattr(context,"showgap"):
            self.showgap = context.showgap
        else:
            self.showgap = 4096  ##  interval between displays
                
        ##  each genotype is a list (at present).
        
    def initpop(self,popsize=20,bias=[]):
        "initializes genepool randomly."
        self.popsize = popsize
        e = self.context  ##  'environment'
        self.genotype = [] ; self.quality = []  ##  clear genepool
        self.best = -Genepool.huge  ##  also reset highest score
        if self.showoff:
            print("SEAL incr", self.incr)
        for i in range(self.popsize):
            x = []
            for j in range(self.glen):
                if bias and i >= 7:  ##  user supplies a list of integers
                    r = random.choice(bias)
                else:  ##  just pick a random number
                    r = random.randint(0,self.gmax)
                x.append(r)
            self.genotype.append(x)
            ##  do fitness evaluation as well:
            q = self.fitness(x,e) ; self.quality.append(q)
            if q > self.best:  ##  fitness must maximize
                self.best = q ; self.bestitem = copy(x)
                ##  later may need deepcopy() rather than copy().
        ##  eventually user should provide function to make a genome.

    def showpop(self):
        "basic genepool display routine."
        print(self.name, self.popsize, self.glen, self.best)
        for j in range(self.popsize):
            print (j,self.genotype[j])
            print (self.quality[j])
        ##  print("**", self.best, self.bestitem)
        self.showbest()

    def showbest (self):
        "just shows current best item in population."
        print("**",self.best,'\n',self.bestitem)
        
    def mate(this,p1,p2,px):
        "create new offspring from p1 & p2 in px."
        ##  Uniform crossover; p1, p2 & px are pointers (indices):
        for j in range(this.glen):
            r = random.randint(0,144)
            if r % 2 == 0:  ## even number
                g = this.genotype[p2][j]
            else:  ##  odd number
                g = this.genotype[p1][j]
            this.genotype[px][j] = g

    def mutation(self,px):
        "basic mutation method."
        for j in range(self.glen):
            r = random.random() * 100.0
            if self.mu >= r:
                rx = random.randint(-self.incr,self.incr)
                ##  small random increment/decrement.
                gx = rx + self.genotype[px][j]
                if rx == 0 or gx > self.gmax or gx < 0:
                  gx = random.randint(0,self.gmax)
                  ##  if outside range or no-change, just pick new random no.
                self.genotype[px][j] = gx
            if self.swapsize > 0:
                r = random.random() * 100.0
                if self.mu >= r:
                    p1 = random.randint(0,self.glen-1)
                    p2 = random.randint(0,self.glen-1)
                    gx = self.genotype[px][p1]
                    self.genotype[px][p1] = self.genotype[px][p2]
                    self.genotype[px][p2] = gx

    def flea(self,maxtries=256):
        "Forsyth's Little Evolutionary Algorithm."
        t = self.popsize  ##  popsize trials already done.
        m2 = maxtries // 2  ##  halfway
        q1 = maxtries // 4  ##  first quartile
        q3 = maxtries - q1  ##  third quartile
        self.mu = self.mu * 2.0
        while t < maxtries:
            p1 = random.randint(1,self.popsize) - 1
            p2 = random.randint(1,self.popsize) - 1
            px = random.randint(0,self.popsize-1)
            if self.quality[px] > self.quality[p1]:
                px,p1 = p1,px
            if self.quality[px] > self.quality[p2]:
                px,p2 = p2,px
            ##  item at px is now least of 3; site of offspring.
            self.mate(p1,p2,px)
            t = t + 1
            ##  do some mutation, occasionally:
            self.mutation(px)
            ##  print self.genotype[px]
            e = self.fitness(self.genotype[px],self.context)
            self.quality[px] = e
            if e > self.best:
                self.best = e ; self.bestitem = copy(self.genotype[px])
                if self.showoff > 1:
                    print(t,round(e,7),'!')
            ##  if hasattr(self,"topshow") and self.ownshow < 1:
            if t in [self.popsize+1,q1,m2,q3] and hasattr(self,"topshow"):
                self.topshow(self.context,self.bestitem)
                self.lastshow = t
            if t in [q1,q3]:  ##  (in/de-)crease mupc in 2 stages
                self.mu = self.mu * 0.5
            if t in [self.popsize+1,q1,m2,q3,maxtries] and self.showoff == 1:
                print(t,round(self.best,5),'!')
        if self.showoff:
            print(maxtries,"trials made.")
            
##            if t <= (self.popsize + self.zonksize):
##                self.mutator1(px)  ##  fills self.zonklist
##            else:
##                self.mutator2(px)  ##  uses self.zonklist
                
    def mutator1(self,px):
        "basic mutation method."
        zonx = 0
        for j in range(self.glen):
            r = random.random() * 100.0
            if self.mu >= r:
                zonx += 1
                rx = random.randint(-self.incr,self.incr)
                ##  small random increment/decrement.
                gx = rx + self.genotype[px][j]
                if rx == 0 or gx > self.gmax or gx < 0:
                  gx = random.randint(0,self.gmax)
                  ##  if outside range or no-change, just pick new random no.
                self.genotype[px][j] = gx
            if self.swapsize > 0:  ##  also might do a gene-swap
                r = random.random() * 100.0
                if self.mu >= r:
                    zonx += 1
                    p1 = random.randint(0,self.glen-1)
                    p2 = random.randint(0,self.glen-1)
                    gx = self.genotype[px][p1]
                    self.genotype[px][p1] = self.genotype[px][p2]
                    self.genotype[px][p2] = gx
        if len(self.zonklist) < Genepool.zonksize:
            self.zonklist.append(zonx)

    def mutator2(self,px):
        "secondary (faster?) mutation method."
        zonx = random.choice(self.zonklist)
        for i in range(zonx):
            r = random.random()
            if r > 0.5 or self.swapsize <= 0:
                j = random.randint(0,self.glen-1)
                rx = random.randint(-self.incr,self.incr)
                ##  small random increment/decrement.
                gx = rx + self.genotype[px][j]
                if rx == 0 or gx > self.gmax or gx < 0:
                  gx = random.randint(0,self.gmax)
                  ##  if outside range or no-change, just pick new random no.
                self.genotype[px][j] = gx
            if self.swapsize > 0 and r <= 0.5:  ##  swap 2 genes
                p1 = random.randint(0,self.glen-1)
                p2 = random.randint(0,self.glen-1)
                gx = self.genotype[px][p1]
                self.genotype[px][p1] = self.genotype[px][p2]
                self.genotype[px][p2] = gx


##  Main routine for simple testing:

def main():
    genomes = Genepool("Ecosystem_1",fitness,[],12)
    genomes.initpop(20)
    ##  genomes.showpop()
    genomes.flea(1000)
    genomes.showpop()
    

##  Call main and attempt to honour exit status:
if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        sys.exit(1)
